@@ -4,10 +4,10 @@ Todo:
     * Implement Anchors
 """
 
+import numpy as np
 import math
 from typing import Callable, Optional, Sequence, Tuple, Union
 
-import numpy as np
 from instancelib import (AbstractEnvironment, Instance, InstanceProvider,
                          LabelProvider, MemoryLabelProvider, TextEnvironment)
 from instancelib.machinelearning import AbstractClassifier
@@ -21,12 +21,19 @@ from text_explainability.data.weights import (exponential_kernel,
                                               pairwise_distances)
 from text_explainability.default import Readable
 from text_explainability.generation.feature_selection import FeatureSelector
-from text_explainability.generation.return_types import FeatureAttribution
+from text_explainability.generation.return_types import FeatureAttribution, Rules
 from text_explainability.generation.surrogate import (LinearSurrogate,
-                                                      TreeSurrogate)
+                                                      TreeSurrogate,
+                                                      RuleSurrogate)
 from text_explainability.generation.target_encoding import FactFoilEncoder
 from text_explainability.utils import binarize, default_detokenizer
 
+import six
+import sys
+sys.modules['sklearn.externals.six'] = six  # ensure backward compatibility
+
+from skrules import SkopeRules  # noqa: E402
+
 
 def default_env(env: Optional[AbstractEnvironment] = None) -> AbstractEnvironment:
     """If no environment is supplied, an empty Enviroment is created for text data.
@@ -132,7 +139,7 @@ class LocalExplanation(Readable):
 
         if predict:
             return provider, perturbed, y
-        return provider, perturbed
+        return provider, perturbed        
 
 
 class WeightedExplanation:
@@ -262,9 +269,9 @@ class LIME(LocalExplanation, WeightedExplanation):
             feature_importances.append(self.local_model.feature_importances)
             used_features[label] = features
 
-        return FeatureAttribution(provider,
-                                  used_features,
-                                  feature_importances,
+        return FeatureAttribution(provider=provider,
+                                  scores=feature_importances,
+                                  used_features=used_features,
                                   labels=labels,
                                   labelset=self.labelset)
 
@@ -388,11 +395,11 @@ class KernelSHAP(LocalExplanation):
                 except np.linalg.LinAlgError:
                     tmp2 = np.linalg.pinv(np.dot(X_W, X))
                 phi = np.dot(tmp2, np.dot(X_W, y)).T
-        return FeatureAttribution(provider,
-                                  used_features,
+        return FeatureAttribution(provider=provider,
                                   scores=phi,
                                   scores_stddev=phi_var,
                                   base_score=y_null,
+                                  used_features=used_features,
                                   labels=np.arange(y.shape[1]),
                                   labelset=self.labelset)
 
@@ -442,17 +449,18 @@ class Anchor(LocalExplanation):
         assert 0.0 <= epsilon <= 0.95, f'epsilon should be a value in [0, 1], but is {epsilon}'
         assert batch_size > 2, 'requires positive batch size'
 
-        y = [provider[i] for i in range(batch_size + 1)]
-        y_true, y = y[0], y[1:]  # noqa: F841
+        for batch in provider.instance_chunker(batch_size):
+            y = list(model.predict_proba_raw(batch))[-1][-1]  # todo: only look at probs of one class
+            y_true, y = y[0], y[1:]  # noqa: F841
 
-        beta = np.log(1.0 / delta)
-        mean = y.mean()
-        lb = Anchor.dlow_bernoulli(mean, beta / perturbed.shape[0])
+            beta = np.log(1.0 / delta)
+            mean = y.mean()
+            lb = Anchor.dlow_bernoulli(mean, beta / perturbed.shape[0])
 
-        batch = 1
-        while mean > min_confidence and lb < min_confidence - epsilon:
-            batch += 1
-        pass
+            if not(mean > min_confidence and lb < min_confidence - epsilon):
+                break
+
+        raise NotImplementedError('[WIP] Implementing anchor/anchor_base.py')
 
     def __call__(self,
                  sample: MemoryTextInstance,
@@ -470,8 +478,7 @@ class Anchor(LocalExplanation):
                                                   contiguous=False, n_samples=n_samples,
                                                   predict=False)
         perturbed = binarize(perturbed[1:])  # flatten all n replacements into one
-        y_true = model(provider[0])
-        y_true = np.argmax(y_true) if y_true.ndim > 1 else y_true
+        y_true = np.argmax(model.predict_proba([provider[0]])[0][-1])  # noqa: F841
 
         # Use beam from https://homes.cs.washington.edu/~marcotcr/aaai18.pdf (Algorithm 2)
         anchor = Anchor.beam_search(provider,  # noqa: F841
@@ -499,7 +506,7 @@ class LocalTree(LocalExplanation, WeightedExplanation):
         LocalExplanation.__init__(self, env=env, augmenter=augmenter, labelset=labelset, seed=seed)
         WeightedExplanation.__init__(self, kernel=kernel, kernel_width=kernel_width)
         if local_model is None:
-            local_model = TreeSurrogate(DecisionTreeClassifier(max_depth=3))
+            local_model = TreeSurrogate(DecisionTreeClassifier(random_state=self._seed))
         self.local_model = local_model
         self.explanation_type = explanation_type
 
@@ -509,33 +516,98 @@ class LocalTree(LocalExplanation, WeightedExplanation):
                  n_samples: int = 50,
                  weigh_samples: bool = True,
                  distance_metric: str = 'cosine',
+                 max_rule_size: Optional[int] = None,
                  **sample_kwargs):
-        _, perturbed, y = self.augment_sample(sample,
-                                              model,
-                                              n_samples=n_samples,
-                                              avoid_proba=True,
-                                              **sample_kwargs)
+        provider, perturbed, y = self.augment_sample(sample,
+                                                     model,
+                                                     n_samples=n_samples,
+                                                     avoid_proba=True,
+                                                     **sample_kwargs)
         perturbed = binarize(perturbed)  # flatten all n replacements into one
 
-        # Sample weights?
         weights = self.weigh_samples(perturbed, metric=distance_metric) if weigh_samples else None
+        self.local_model.max_rule_size = max_rule_size
         self.local_model.fit(perturbed, y, weights=weights)
 
-        return self.local_model.feature_importances, self.local_model
+        return Rules(provider=provider,
+                     rules=self.local_model,
+                     labelset=self.labelset,
+                     sampled=True)
 
 
-class FoilTree(LocalExplanation, WeightedExplanation):
+class FactFoilMixin:
+    def to_fact_foil(self, y, labelset, foil_fn: Union[FactFoilEncoder, int, str]):
+        if isinstance(foil_fn, str):
+            foil_fn = FactFoilEncoder.from_str(foil_fn, labelset)
+        elif isinstance(foil_fn, int):
+            foil_fn = FactFoilEncoder(foil_fn, labelset)
+        return foil_fn(y)
+
+
+class FoilTree(FactFoilMixin, LocalExplanation, WeightedExplanation):
     def __init__(self,
                  env: Optional[AbstractEnvironment] = None,
                  labelset: Optional[Union[Sequence[str], LabelProvider]] = None,
                  augmenter: Optional[LocalTokenPertubator] = None,
-                 local_model: Optional[TreeSurrogate] = DecisionTreeClassifier(),
+                 local_model: Optional[TreeSurrogate] = None,
                  kernel: Optional[Callable] = None,
                  kernel_width: Union[int, float] = 25,
                  explanation_type: str = 'multiclass',
                  seed: int = 0):
         LocalExplanation.__init__(self, env=env, augmenter=augmenter, labelset=labelset, seed=seed)
         WeightedExplanation.__init__(self, kernel=kernel, kernel_width=kernel_width)
+        if local_model is None:
+            local_model = TreeSurrogate(DecisionTreeClassifier(random_state=self._seed))
+        self.local_model = local_model
+        self.explanation_type = explanation_type
+
+    def __call__(self,
+                 sample: MemoryTextInstance,
+                 model: AbstractClassifier,
+                 foil_fn: Union[FactFoilEncoder, int, str],
+                 n_samples: int = 50,
+                 weigh_samples: bool = True,
+                 distance_metric: str = 'cosine',
+                 max_rule_size: Optional[int] = None,
+                 **sample_kwargs):
+        provider, perturbed, y = self.augment_sample(sample,
+                                                     model,
+                                                     n_samples=n_samples,
+                                                     avoid_proba=True,
+                                                     **sample_kwargs)
+        perturbed = binarize(perturbed)  # flatten all n replacements into one
+
+        # Encode foil as 0 and rest as 1
+        labelset = self.labelset if self.labelset else model
+        y_ = self.to_fact_foil(y, labelset, foil_fn)
+
+        weights = self.weigh_samples(perturbed, metric=distance_metric) if weigh_samples else None
+        self.local_model.max_rule_size = max_rule_size
+        self.local_model.fit(perturbed, y_, weights=weights)
+
+        # TODO: pass to which label the Foil Tree applies
+        return Rules(provider=provider,
+                     rules=self.local_model,
+                     labelset=labelset,
+                     sampled=True)
+
+
+class LocalRules(FactFoilMixin, LocalExplanation, WeightedExplanation):
+    def __init__(self,
+                 env: Optional[AbstractEnvironment] = None,
+                 labelset: Optional[Union[Sequence[str], LabelProvider]] = None,
+                 augmenter: Optional[LocalTokenPertubator] = None,
+                 local_model: Optional[RuleSurrogate] = None,
+                 kernel: Optional[Callable] = None,
+                 kernel_width: Union[int, float] = 25,
+                 explanation_type: str = 'multiclass',
+                 seed: int = 0):
+        LocalExplanation.__init__(self, env=env, augmenter=augmenter, labelset=labelset, seed=seed)
+        WeightedExplanation.__init__(self, kernel=kernel, kernel_width=kernel_width)
+        if local_model is None:
+            local_model = RuleSurrogate(SkopeRules(max_depth_duplication=2,
+                                                   n_estimators=30,
+                                                   random_state=self._seed))
         self.local_model = local_model
         self.explanation_type = explanation_type
 
@@ -547,24 +619,21 @@ class FoilTree(LocalExplanation, WeightedExplanation):
                  weigh_samples: bool = True,
                  distance_metric: str = 'cosine',
                  **sample_kwargs):
-        _, perturbed, y = self.augment_sample(sample,
-                                              model,
-                                              n_samples=n_samples,
-                                              avoid_proba=True,
-                                              **sample_kwargs)
+        provider, perturbed, y = self.augment_sample(sample,
+                                                     model,
+                                                     n_samples=n_samples,
+                                                     avoid_proba=True,
+                                                     **sample_kwargs)
+        perturbed = binarize(perturbed)  # flatten all n replacements into one
 
         # Encode foil as 0 and rest as 1
         labelset = self.labelset if self.labelset else model
-        if isinstance(foil_fn, str):
-            foil_fn = FactFoilEncoder.from_str(foil_fn, labelset)
-        elif isinstance(foil_fn, int):
-            foil_fn = FactFoilEncoder(foil_fn, labelset)
-        y_ = foil_fn(y)
+        y_ = self.to_fact_foil(y, labelset, foil_fn)
 
-        perturbed = binarize(perturbed)  # flatten all n replacements into one
+        weights = self.weigh_samples(perturbed, metric=distance_metric) if weigh_samples else None
+        self.local_model.fit(perturbed, y_, weights=weights)
 
-        # weights = self.weigh_samples(perturbed, metric=distance_metric) if weigh_samples else None
-        # self.local_model.fit(perturbed, y_, weights=weights)
-        self.local_model.fit(perturbed, y_)
-
-        return self.local_model, self.local_model.tree_.max_depth
+        return Rules(provider,
+                     rules=self.local_model,
+                     labelset=labelset,
+                     sampled=True)
